#include <stdio.h>
#include <termios.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>

#define ESC "\033"
#define CSI ESC "["
#define CSIQ CSI "?"

void cmd0(const char *code)
{
  printf("%s",code);
  fflush(stdout);
}

void cmd1(const char *code, int arg, const char *end)
{
  if (end) printf("%s%d%s", code, arg, end);
     else  printf("%s%d", code, arg);
  fflush(stdout);
}

void cmd2(const char *code, int arg1, int arg2, const char *end)
{
  if (end) printf("%s%d;%d%s",code, arg1,arg2, end);
      else printf("%s%d;%d",code, arg1,arg2);
  fflush(stdout);
}

void altbuf_enter() { cmd1(CSIQ, 1049, "h"); }
void altbuf_leave() { cmd1(CSIQ, 1049, "l"); }
void enable_left_right_margins() { cmd1(CSIQ, 69, "h"); }
void disable_left_right_margins() { cmd1(CSIQ, 69, "l"); }
void enable_DECPFF() { cmd1(CSIQ, 18, "h"); }
void disable_DECPFF() { cmd1(CSIQ, 18, "l"); }

void enable_printer_only_scroll_region() { cmd1(CSIQ, 19, "h"); }
void disable_printer_only_scroll_region() { cmd1(CSIQ, 19, "l"); }

void enable_wraparound() { cmd1(CSIQ, 7, "h"); }
void disable_wraparound() { cmd1(CSIQ, 7, "l"); }

void cDECBI() { cmd0( ESC "6" ); }
void cDECSC() { cmd0( ESC "7" ); }
void cDECRC() { cmd0( ESC "8" ); }
void cDECFI() { cmd0( ESC "9" ); }
void cIND() { cmd0( ESC "D" ); }
void cNEL() { cmd0( ESC "E" ); }
void HP_home_down() { cmd0( ESC "F" ); }
// ESC G is rxvt graphics
void cHTS() { cmd0(ESC "H"); } 
void cRI() { cmd0( ESC "M" ); }
// ESC N and ESC O are for old school character encodings
void cSPA() { cmd0( ESC "V" ); }
void cEPA() { cmd0( ESC "W" ); }
void cDECID() { cmd0( ESC "Z" ); }
void cRIS() { cmd0( ESC "c" ); }
void HP_memory_lock() { cmd0(ESC "l"); }
void HP_memory_unlock() { cmd0(ESC "m"); }
// ESC n and ESC o are for old school character encodings
void cDECKPNM() { cmd0(ESC ">"); }
void cDECKPAM() { cmd0(ESC "="); }
// ESC |  }  ~ are for old school character encodings
// ESC # 3 4 5 6 are for double-width/height stuff
void cDECALN() { cmd0(ESC "#8" ); }
void enable_utf8() { cmd0(ESC "%8"); }
void disable_utf8() { cmd0(ESC "%@"); }
void seven_bit_controls() { cmd0( ESC " F"); }
void eight_bit_controls() { cmd0( ESC " G"); }
// ESC <space> L M N are for old school character encodings

void cCUU(int n) { cmd1(CSI, n, "A"); }
void cSR(int n) { cmd1(CSI, n, " A"); }
void cCUD(int n) { cmd1(CSI, n, "B"); }
void cCUF(int n) { cmd1(CSI, n, "C"); }
void cCUB(int n) { cmd1(CSI, n, "D"); }
void cCNL(int n) { cmd1(CSI, n, "E"); }
void cCPL(int n) { cmd1(CSI, n, "F"); }
void cHPA(int n) { cmd1(CSI, n, "G"); }
void cCUP(int r, int c) { cmd2(CSI, r, c, "H"); }
void cCHT(int n) { cmd1(CSI, n, "I"); }
void cED(int n) { cmd1(CSI, n, "J"); }
void cDECSED(int n) { cmd1(CSIQ, n, "J"); }
void cEL(int n) { cmd1(CSI, n, "K"); }
void cDECSEL(int n) { cmd1(CSIQ, n, "K"); }
void cIL(int n) { cmd1(CSI, n, "L"); }
void cDL(int n) { cmd1(CSI, n, "M"); }
void cDCH(int n) { cmd1(CSI, n, "P"); }
void cSU(int n) { cmd1(CSI, n, "S"); }
void cSD(int n) { cmd1(CSI, n, "T"); }
// there are other stuff such as title mode, mouse tracking etc.
// those are for later, if ever.
void cursor_tabulation_control(int n) { cmd1(CSI, n, "W"); }
void set_tab_at_every_8() { cmd0(CSIQ "W"); }
void cECH(int n) { cmd1(CSI, n, "X"); }
void cCBT(int n) { cmd1(CSI, n, "Z"); }
void cHPR(int n) { cmd1(CSI, n, "a"); }
void cREP(int n) { cmd1(CSI, n, "b"); }
//
// CSI c is device attributes and linux cursor style.
// delayed
void cVPA(int n) { cmd1(CSI, n, "d"); }
void unset_tab_stop(int n) { cmd1(CSI, n, " d"); }
void cVPR(int n) { cmd1(CSI, n, "e"); }
// CSI f is alias to CUP
void cTBC(int n) { cmd1(CSI, n, "g"); }
// CSI h is SETMODE,
// CSI ? h is also SETMODE,
// handled elsewhere
// CSI i is Media Control, which is attached line printer stuff.
// 
// CSI j and CSI k are used by urxvt as aliases to other things
// CSI space k exists, but is irrelevant
// CSI l and CSI ?l reset modes, undoing what SETMODE does
//
void cSGR(int n) { cmd1(CSI, n, "m"); }
// There are others starting with CSI and ending with m, but
// I won't do those, they are mostly xterm specific.
//
// CSI n reports some status..
//
void cDECSTR() { cmd0(CSI "!p"); }
void cDECSTBM(int a1, int a2) { cmd2(CSI, a1, a2, "r"); }
void cDECSLRM(int a1, int a2) { cmd2(CSI, a1, a2, "s"); }
//
// There are a lot more xterm specific sequences, but I won't
// document them all, this is getting boring..
//
void cICH(int n) { cmd1(CSI, n, "@"); }

void O_goto(int r,int c) { cCUP(r,c); }





static struct termios termio_settings;

void resetraw()
{
  tcsetattr(0, TCSANOW, &termio_settings);
  altbuf_leave();
  fflush(stdout);
}

void makeraw()
{
  struct termios tt;
  tcgetattr(0, &termio_settings);
  tt= termio_settings;
  cfmakeraw(&tt);
  atexit(resetraw);
  tcsetattr(0, TCSANOW, &tt);
}


int getK()
{
  char b[1];
  b[0]= 0;
  read(0, b, 1);
  return b[0];
}


int dump_window()
{
  cDECSTBM( 1, 24);
  cmd0( CSI "0i" );
}

int tt; // term type 'x'=xterm 'm'=muterm



void mudebug() { if (tt=='m') cmd0( ESC "0" ); }


void patV()
{
  int i,j;
  for(i=0;i<24;i++)
  {
    O_goto(i+1, 1);
    if (i!=12)
    for(j=0;j<80;j++) putc('0'+i%10,stdout);
    else
    for(j=0;j<40;j++) putc('0'+i%10,stdout);
  }
}

static void test_IND_1()
{
  patV();
  cDECSTBM(3, 14);
  O_goto(14, 1);
  mudebug();
  cIND();
}

static void test_IND_2()
{
  patV();
  cDECSTBM(3, 14);
  enable_left_right_margins();
  cDECSLRM(4, 21);
  O_goto(14, 6);
  mudebug();
  cIND();
}

static void test_BI_1()
{
  patV();
  cDECSTBM(3, 14);
  enable_left_right_margins();
  cDECSLRM(4, 21);
  O_goto(12, 4);
  mudebug();
  cDECBI();
}

static void test_BI_2()
{
  patV();
  cDECSTBM(3, 14);
  enable_left_right_margins();
  cDECSLRM(4, 21);
  O_goto(12, 5);
  mudebug();
  cDECBI();
}

static struct
{
  int number;
  void (*func)();
} tests[]= 
{
  { 100, test_IND_1 },
  { 101, test_IND_2 },
  { 102, test_BI_1 },
  { 103, test_BI_2 },
  { -1,  NULL }
};

int main(int argc,char **argv)
{
  int i;
  int tn;
  int wt;
  // int all_tests;

  tn= 0;
  wt= 0;
  // all_tests= 0;
  for(i=1;i<argc;i++)
  {
    if (!strcmp(argv[i],"-m")) { tt= 'm'; continue; }
    if (!strcmp(argv[i],"-x")) { tt= 'x'; continue; }
    if (!strcmp(argv[i], "-w")) { wt= 1; continue; }
  //  if (!strcmp(argv[i],"-a")) { all_tests= 1; continue; }
    tn= atoi(argv[i]);
  }

  makeraw();
  altbuf_enter();
  disable_printer_only_scroll_region();
  enable_DECPFF();

  for(i=0;tests[i].func;i++)
   if (tn==tests[i].number)
   {
     (*tests[i].func)();
     fflush(stdout);
     dump_window();
     break;
   }

  while(wt)
  {
    switch(getK())
    {
    case 'q':
    case 'Q':
      wt= 0;
    }
  }
  return 0;
}
